#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.

import json
from typing import Any, Dict, List, Optional, Type
from datetime import datetime, timedelta

from pydantic import BaseModel
from tornado.httpclient import AsyncHTTPClient

from .db import get_user_field
from .schema import UJ_STATIC_USER_ID, UJ_USER_EMAIL
from .telemetry_utils import (
    MITO_SERVER_FREE_TIER_LIMIT_REACHED,
    log,
)
from .version_utils import is_pro

MITO_AI_URL = "https://ogtzairktg.execute-api.us-east-1.amazonaws.com/Prod/completions/"

OPEN_SOURCE_AI_COMPLETIONS_LIMIT = 500
OPEN_SOURCE_INLINE_COMPLETIONS_LIMIT = 30 # days

__user_email = None
__user_id = None


def check_mito_server_quota(n_counts: int, first_usage_date: str) -> None:
    """Check whether the user has reached the limit of completions for the free tier or not.

    Args:
        n_counts: The number of completions the user has made so far.
        first_usage_date: The date of the user's first usage.
    Raises:
        PermissionError: If the user has reached the limit.
    """
    pro = is_pro()

    if pro:
        return

    if n_counts >= OPEN_SOURCE_AI_COMPLETIONS_LIMIT:
        log(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
        raise PermissionError(MITO_SERVER_FREE_TIER_LIMIT_REACHED)

    if first_usage_date != "":
        first_use = datetime.strptime(first_usage_date, "%Y-%m-%d")
        one_month_later = first_use + timedelta(days=OPEN_SOURCE_INLINE_COMPLETIONS_LIMIT)
        if datetime.now() > one_month_later:
            log(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
            raise PermissionError(MITO_SERVER_FREE_TIER_LIMIT_REACHED)


async def get_ai_completion_from_mito_server(
    last_message_content: str,
    ai_completion_data: Dict[str, Any],
    n_counts: int,
    first_usage_date: str,
) -> str:
    global __user_email, __user_id

    if __user_email is None:
        __user_email = get_user_field(UJ_USER_EMAIL)
    if __user_id is None:
        __user_id = get_user_field(UJ_STATIC_USER_ID)

    check_mito_server_quota(n_counts, first_usage_date)

    data = {
        "email": __user_email,
        "user_id": __user_id,
        "data": ai_completion_data,
        "user_input": last_message_content,  # We add this just for logging purposes
    }

    headers = {
        "Content-Type": "application/json",
    }

    http_client = AsyncHTTPClient(defaults=dict(user_agent="Mito-AI client"))
    try:
        res = await http_client.fetch(
            MITO_AI_URL, method="POST", headers=headers, body=json.dumps(data)
        )
    finally:
        http_client.close()

    # The lambda function returns a dictionary with a completion entry in it,
    # so we just return that.
    content = json.loads(res.body)

    return content.get("completion", "")


def get_open_ai_completion_function_params(
    model: str, 
    messages: List[Dict[str, Any]], 
    stream: bool,
    response_format: Optional[Type[BaseModel]] = None
) -> Dict[str, Any]:
    
    completion_function_params = {
        "model": model,
        "stream": stream,
        "messages": messages,
        "response_format": response_format
    }
    
    # o3-mini will error if we try setting the temperature
    if model == "gpt-4o-mini":
        completion_function_params["temperature"] = 0.0

    return completion_function_params


from typing import List

def create_chat_prompt(
    variables: List[str],
    active_cell_code: str, 
    input: str
) -> str:
    variables_str = '\n'.join([f"{variable}" for variable in variables])
    prompt = f"""You are an expert python programmer writing a script in a Jupyter notebook. You are given a set of variables, existing code, and a task.

There are two possible types of responses you might give:
1. Code Update: If the task requires modifying or extending the existing code, respond with the updated active code cell and a short explanation of the changes made. 
2. Explanation/Information: If the task does not require a code update, provide an explanation, additional information about a package, method, or general programming question, without writing any code. Keep your response concise and to the point.

When responding:
- Do not use the word "I"
- Do not recreate variables that already exist
- Keep as much of the original code as possible

<Example>

Defined Variables:
{{
    'loan_multiplier': 1.5,
    'sales_df': pd.DataFrame({{
        'transaction_date': ['2024-01-02', '2024-01-02', '2024-01-02', '2024-01-02', '2024-01-03'],
        'price_per_unit': [10, 9.99, 13.99, 21.00, 100],
        'units_sold': [1, 2, 1, 4, 5],
        'total_price': [10, 19.98, 13.99, 84.00, 500]
    }})
}}

Code in the active code cell:
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')
```

Your task: convert the transaction_date column to datetime and then multiply the total_price column by the sales_multiplier.

Output:
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')
sales_df['transaction_date'] = pd.to_datetime(sales_df['transaction_date'])
sales_df['total_price'] = sales_df['total_price'] * sales_multiplier
```

Converted the `transaction_date` column to datetime using the built-in pd.to_datetime function and multiplied the `total_price` column by the `sales_multiplier` variable.

</Example>

Defined Variables:
{variables_str}

Code in the active code cell:
```python
{active_cell_code}
```

Your task: {input}
"""
    
    return prompt

def create_explain_code_prompt(active_cell_code: str):
    prompt = f"""Explain the code in the active code cell to me like I have a basic understanding of Python. Don't explain each line, but instead explain the overall logic of the code.

<Example>

Code in the active code cell:

```python
def multiply(x, y):
    return x * y
```

Output:

This code creates a function called `multiply` that takes two arguments `x` and `y`, and returns the product of `x` and `y`.

</Example>

Code in the active code cell:

```python
{active_cell_code}
```

Output: 
"""
    return prompt
    
def create_inline_prompt(
        prefix: str,
        suffix: str,
        variables: List[str],
):
    variables_str = '\n'.join([f"{variable}" for variable in variables])
    prompt = f"""You are a coding assistant that lives inside of JupyterLab. Your job is to help the user write code.

You're given the current code cell, the user's cursor position, and the variables defined in the notebook. The user's cursor is signified by the symbol <cursor>.

CRITICAL FORMATTING RULES:
1. Include a new line character at the start of your response if you want the code you are writing to be added on the line after the cursor. For example, if the cursor is at the end of a comment, you should start your response with a newline character so that the code you write is not added to the comment.
2. If you are finishing a line of code that the user started, return the full line of code with no newline character at the start or end.
3. Your response must preserve correct Python indentation and spacing. For example, if you're completing a line of indented code, you must preserve the indentation.

Your job is to complete the code that matches the user's intent. Write the minimal code to achieve the user's intent. Don't expand upon the user's intent.

<Example 1>
Defined Variables: {{
    'loan_multiplier': 1.5,
    'sales_df': pd.DataFrame({{
        'transaction_date': ['2024-01-02', '2024-01-02', '2024-01-02', '2024-01-02', '2024-01-03'],
        'price_per_unit': [10, 9.99, 13.99, 21.00, 100],
        'units_sold': [1, 2, 1, 4, 5],
        'total_price': [10, 19.98, 13.99, 84.00, 500]
    }})
}}

Code in the active code cell:
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')

# Multiply the total_price column by the loan_multiplier<cursor>
```

Output:
```python

sales_df['total_price'] = sales_df['total_price'] * loan_multiplier
```
</Example 1>

IMPORTANT: Notice in Example 1 that the output starts with a newline because the cursor was at the end of a comment. This newline is REQUIRED to maintain proper Python formatting.

<Example 2>
Defined Variables: {{
    df: pd.DataFrame({{
        'age': [20, 25, 22, 23, 29],
        'name': ['Nawaz', 'Aaron', 'Charlie', 'Tamir', 'Eve'],
    }})
}}

Code in the active code cell:
```python
df['age'] = df[<cursor>['age'] > 23]
```

Output:
```python
df['age'] = df[df['age'] > 23]
```
</Example 2>

IMPORTANT: Notice in Example 2 that the output does NOT start with a newline because the cursor is in the middle of existing code.

<Example 3>
Defined Variables: {{}}

Code in the active code cell:
```python
voters = pd.read_csv('./voters.csv')

# Create a variable for pennsylvania voters, ohio voters, california voters, and texas voters
pa_voters = voters[voters['state'] == 'PA']
ohio_voters<cursor>
```

Output:
```python
ohio_voters = voters[voters['state'] == 'OH']
ca_voters = voters[voters['state'] == 'CA']
tx_voters = voters[voters['state'] == 'TX']
```
</Example 3>

IMPORTANT: Notice in Example 3 that output does not start with a newline character because it wasnts to continue the line of code that the user started. Also notice the output contains three lines of code because that is the minimal code to achieve the user's intent.

<Example 4>
Defined Variables: {{}}

Code in the active code cell:
```python
# Display the first 5 rows of the dataframe
df.head()
<cursor>
```

Output:
```python
```
</Example 4>

IMPORTANT: Notice in Example 4 that the output is empty becuase the user's intent is already complete.

<Example 5>
Defined Variables: {{}}

Code in the active code cell:
```python
def even_and_odd():
    for i in range(10):
        if i % 2 == 0:
            print(f"Even: {{i}}")
        else:
            pri<cursor>
```

Output:
```python
            print(f"Odd: {{i}}")
```
</Example 5>

IMPORTANT: Notice in Example 5 that the output is indented several times because the code must be executed as part of the else block.

<Example 6>
Defined Variables: {{}}

Code in the active code cell:
```python
days_in_week <cursor>
```

Output:
```python
days_in_week = 7
```
</Example 6>

IMPORTANT: Notice in Example 6 that inorder to finish the variable declaration, the output continues the existing line of code and does not start with a new line character.

Your Task:

Defined Variables: {variables_str}

Code in the active code cell:
```python
{prefix}<cursor>{suffix}
```

Output:
"""
    return prompt

def create_error_prompt(
    errorMessage: str,
    active_cell_code: str,
    variables: List[str],
):
    variables_str = '\n'.join([f"{variable}" for variable in variables])
    prompt = f"""You are debugging code in a JupyterLab 4 notebook. Analyze the error and provide a solution that maintains the original intent.

<Example 1>
Defined Variables:
{{
    'revenue_multiplier': 1.5,
    'sales_df': pd.DataFrame({{
        'transaction_date': ['2024-01-02', '2024-01-02', '2024-01-02', '2024-01-02', '2024-01-03'],
        'price_per_unit': [10, 9.99, 13.99, 21.00, 100],
        'units_sold': [1, 2, 1, 4, 5],
        'total_price': [10, 19.98, 13.99, 84.00, 500]
    }})
}}

Code in the active code cell:
```python
sales_df['total_revenue'] = sales_df['price'] * revenue_multiplier
```

Error Message:
KeyError: 'price'

ERROR ANALYSIS:
Runtime error: Attempted to access non-existent DataFrame column

INTENT ANALYSIS:
User is trying to calculate total revenue by applying a multiplier to transaction prices. Based on the defined variables, the column that the user is tring to access is likely `total_price` because that would allow them to calculate the total revenue for each transaction.

SOLUTION:
```python
sales_df['total_revenue'] = sales_df['total_price'] * revenue_multiplier
```

The DataFrame contains 'total_price' rather than 'price'. Updated column reference to match existing data structure.
</Example 1>

<Example 2>
Defined Variables:
{{
    'df': pd.DataFrame({{
        'order_id': [1, 2, 3, 4],
        'date': ['Mar 7, 2025', 'Sep 24, 2024', '25 June, 2024', 'June 29, 2024'],
        'amount': [100, 150, 299, 99]
    }})
}}

Code in the active code cell:
```python
df['date'] = pd.to_datetime(df['date'])
```

Error Message:
ValueError: time data "25 June, 2024" doesn't match format "%b %d, %Y", at position 2. You might want to try:
    - passing `format` if your strings have a consistent format;
    - passing `format='ISO8601'` if your strings are all ISO8601 but not necessarily in exactly the same format;
    - passing `format='mixed'`, and the format will be inferred for each element individually. You might want to use `dayfirst` alongside this.

ERROR ANALYSIS:
This is a ValueError caused by applying the wrong format to a specific date string. Because it was triggered at position 2, the first date string must have successfully converted. By looking at the defined variables, I can see that first date string is in the format "Mar 7, 2025", but the third date string is in the format "25 June, 2024". Those dates are not in the same format, so the conversion failed.

INTENT ANALYSIS:
User is trying to convert the date column to a datetime object even though the dates are not in the same starting format. 

SOLUTION:
```python
def parse_date(date_str):
    formats = ['%b %d, %Y', '%d %B, %Y']
    
    for fmt in formats:
        try:
            return pd.to_datetime(date_str, format=fmt)
        except ValueError:
            # Try the next format
            continue
            
    # If no format worked, return Not a Time
    return pd.NaT

df['date'] = df['date'].apply(lambda x: parse_date(x))
```

Since the dates are not in a consistent format, we need to first figure out which format to use for each date string and then use that format to convert the date.

The best way to do this is with a function. We can call this function `parse_date`.
</Example 2>


Guidelines for Solutions:

Error Analysis:

- Identify error type (Syntax, Runtime, Logic).
- Use the defined variables and code in the active cell to understand the error.
- Consider kernel state and execution order

Intent Preservation:

- Try to understand the user's intent using the defined variables and code in the active cell.

Solution Requirements:

- Return the full code cell with the error fixed and a short explanation of the error.
- Propose a solution that fixes the error and does not change the user's intent.
- Make the solution as simple as possible.
- Reuse as much of the existing code as possible.
- Do not add temporary comments like '# Fixed the typo here' or '# Added this line to fix the error'
- The code in the SOLUTION section should be a python code block starting with ```python and ending with ```

Here is your task. 

Defined Variables:
{variables_str}

Code in the active code cell:
```python
{active_cell_code}
```

Error Message:
{errorMessage}

ERROR ANALYSIS:

INTENT ANALYSIS:

SOLUTION:
"""
    return prompt

def create_agent_prompt(file_type: str, columnSamples: List[str], input: str):
    if file_type:
        file_sample_snippet = f"""You will be working with the following dataset (sample rows shown) from a {file_type} file:
{columnSamples}
"""

    return f"""You are an expert data science assistant working in a Jupyter notebook environment. Your task is to break a problem into the essential, actionable steps required to write Python code for solving it. 
You have access to the following Python packages:
- pandas (for data manipulation and analysis)
- matplotlib (for data visualization)
{file_sample_snippet if file_type else ''}
Given the dataset (if provided) and the question below:
1. Break the problem into the **smallest possible number of clear, high-level tasks** necessary to achieve the solution. 
2. **Do not include any code or specific implementation details.** Focus only on describing the high-level steps required to solve the problem.
3. Additionally, provide a list of python packages that are required to complete the actions. 
{input}
"""


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
import json
import os
from pathlib import Path

"""
This code is responsible for getting the package.json that is bundled
with the javascript code bundle inside this Python package, so we can 
get the version of the package. This means we can easily bump the package
version in one place (package.json) and have it automatically update here too. 
This is nice for the release process. 

Since this is a Jupyter Lab 4 extension, the mito-ai/labextension folder
contains the package.json directly. So we can just read it there. 
"""
lab_extension_folder = os.path.join(Path(__file__).parent, 'labextension')

package_json_path = os.path.join(lab_extension_folder, 'package.json')
package_json = json.loads(open(package_json_path).read())

__version__ = package_json['version']


from __future__ import annotations

import os
from typing import Any, AsyncGenerator, Dict, List, Optional, Union, Type

import openai
from openai._streaming import AsyncStream
from openai.types.chat import ChatCompletionChunk
from traitlets import  Instance, Unicode, default, validate, List
from pydantic import BaseModel
from traitlets.config import LoggingConfigurable

from mito_ai.logger import get_logger
from mito_ai.models import (
    AICapabilities,
    CompletionError,
    CompletionItem,
    CompletionItemError,
    CompletionReply,
    CompletionRequest,
    CompletionStreamChunk,
)
from mito_ai.utils.db import get_user_field, set_user_field
from mito_ai.utils.open_ai_utils import (
    check_mito_server_quota,
    get_ai_completion_from_mito_server,
    get_open_ai_completion_function_params,
)

from mito_ai.utils.schema import UJ_AI_MITO_API_NUM_USAGES, UJ_MITO_AI_FIRST_USAGE_DATE
from mito_ai.utils.telemetry_utils import (
    KEY_TYPE_PARAM,
    MITO_AI_COMPLETION_ERROR,
    MITO_SERVER_KEY,
    USER_KEY,
    log,
    log_ai_completion_success,
)

__all__ = ["OpenAIProvider"]
_num_usages = None
_first_usage_date = None

class OpenAIProvider(LoggingConfigurable):
    """Provide AI feature through OpenAI services."""

    api_key = Unicode(
        config=True,
        help="OpenAI API key. Default value is read from the OPENAI_API_KEY environment variable.",
    )
    
    models = List(['gpt-4o-mini', 'o3-mini'])
    
    last_error = Instance(
        CompletionError,
        allow_none=True,
        help="""Last error encountered when using the OpenAI provider.

This attribute is observed by the websocket provider to push the error to the client.""",
    )

    def __init__(self, **kwargs) -> None:
        super().__init__(log=get_logger(), **kwargs)
        self.last_error = None
        self._async_client: Optional[openai.AsyncOpenAI] = None
        self._sync_client: Optional[openai.OpenAI] = None
        self._models: Optional[List[str]] = None

    @default("api_key")
    def _api_key_default(self):
        default_key = os.environ.get("OPENAI_API_KEY", "")
        return self._validate_api_key({"value": default_key})

    @validate("api_key")
    def _validate_api_key(self, changes: Dict[str, Any]) -> str:
        """"""
        api_key = changes["value"]
        if not api_key:
            self.log.debug(
                "No OpenAI API key provided; following back to Mito server API."
            )
            return ""

        client = openai.OpenAI(api_key=api_key)
        models = []
        try:
            # Make an http request to OpenAI to get the models available
            # for this API key.
            # And then handle the exceptions if they are thrown.
            for model in client.models.list():
                models.append(model.id)
            self._models = models
        except openai.AuthenticationError as e:
            self.log.warning(
                "Invalid OpenAI API key provided.",
                exc_info=e,
            )
            self.last_error = CompletionError.from_exception(
                e,
                hint="You're missing the OPENAI_API_KEY environment variable. Run the following code in your terminal to set the environment variable and then relaunch the jupyter server `export OPENAI_API_KEY=<your-api-key>`",
            )
            return ""
        except openai.PermissionDeniedError as e:
            self.log.warning(
                "Invalid OpenAI API key provided.",
                exc_info=e,
            )
            self.last_error = CompletionError.from_exception(e)
            return ""
        except openai.InternalServerError as e:
            self.log.debug(
                "Unable to get OpenAI models due to OpenAI error.", exc_info=e
            )
            return api_key
        except openai.RateLimitError as e:
            self.log.debug(
                "Unable to get OpenAI models due to rate limit error.", exc_info=e
            )
            return api_key
        except openai.APIConnectionError as e:
            self.log.warning(
                "Unable to connect to OpenAI API.",
                exec_info=e,
            )
            self.last_error = CompletionError.from_exception(e)
            return ""
        else:
            self.log.debug("User OpenAI API key validated.")
            return api_key

    @property
    def can_stream(self) -> bool:
        """Whether the provider supports streaming completions.

        Streaming is only supported if an OpenAI API key is provided.
        """
        return bool(self.api_key)

    @property
    def capabilities(self) -> AICapabilities:
        """Get the provider capabilities.

        Returns:
            The provider capabilities.
        """
        if self._models is None:
            self._validate_api_key({"value": self.api_key})

        # If the user has an OpenAI API key, then we don't need to check the Mito server quota.
        if self.api_key:
            return AICapabilities(
                configuration={
                    "model": self.models,
                },
                provider="OpenAI (user key)",
            )

        # Get the number of usages
        global _num_usages
        if _num_usages is None:
            _num_usages = get_user_field(UJ_AI_MITO_API_NUM_USAGES)

        # Get the first usage date
        global _first_usage_date
        _first_usage_date = get_user_field(UJ_MITO_AI_FIRST_USAGE_DATE)
        if _first_usage_date is None:
            from datetime import datetime
            today = datetime.today().strftime('%Y-%m-%d')
            try:
                set_user_field(UJ_MITO_AI_FIRST_USAGE_DATE, today)
            except Exception as e:
                self.log.warning("Failed to set first usage date in user.json", exc_info=e)

        try:
            check_mito_server_quota(_num_usages or 0, _first_usage_date or "")
        except Exception as e:
            self.last_error = CompletionError.from_exception(e)

        return AICapabilities(
            configuration={
                "model": self.models,
            },
            provider="Mito server",
        )

    @property
    def _openAI_async_client(self) -> Optional[openai.AsyncOpenAI]:
        """Get the asynchronous OpenAI client."""
        if not self.api_key:
            return None

        if not self._async_client or self._async_client.is_closed():
            self._async_client = openai.AsyncOpenAI(api_key=self.api_key)

        return self._async_client
    
    @property
    def _openAI_sync_client(self) -> Optional[openai.OpenAI]:
        """Get the synchronous OpenAI client."""
        if not self.api_key:
            return None

        if not self._sync_client or self._sync_client.is_closed():
            self._sync_client = openai.OpenAI(api_key=self.api_key)
            
        return self._sync_client

    async def request_completions(
        self,
        request: CompletionRequest,
        prompt_type: str,
        model: str,
        response_format: Optional[Type[BaseModel]] = None
    ) -> CompletionReply:
        """Get a completion from the OpenAI API.

        Args:
            request: The completion request description.
            prompt_type: The type of prompt that was sent to the AI (e.g. "chat", "smart_debug", "explain")
        Returns:
            The completion
        """
        self.last_error = None
        try:
            if self._openAI_sync_client:
                self.log.debug(f"Requesting completion from OpenAI API with personal key with model: {model}")
                
                # Validate that the model is supported. If not fall back to gpt-4o-mini
                if model not in self.models:
                    model = "gpt-4o-mini"

                completion_function_params = get_open_ai_completion_function_params(model, request.messages, False, response_format)
                completion = self._openAI_sync_client.chat.completions.create(**completion_function_params)
                                
                if prompt_type == "agent:planning":
                    pass # TODO: Add logging for agents 
                else:
                    # Log the successful completion
                    log_ai_completion_success(
                        key_type=USER_KEY,
                        prompt_type=prompt_type,
                        last_message_content=str(request.messages[-1].get('content', '')),
                        response={"completion": completion.choices[0].message.content}
                    )

                return CompletionReply(
                    parent_id=request.message_id,
                    items=[
                        CompletionItem(
                            content=completion.choices[0].message.content or "",
                            isIncomplete=False,
                        )
                    ]
                )
            else:
                # If they don't have an Open AI key, use the mito server to get a completion
                self.log.debug(f"Requesting completion from Mito server with model {model}.")
                global _num_usages
                if _num_usages is None:
                    _num_usages = get_user_field(UJ_AI_MITO_API_NUM_USAGES)
                
                completion_function_params = get_open_ai_completion_function_params(model, request.messages, False, response_format)
                ai_response = await get_ai_completion_from_mito_server(
                    request.messages[-1].get("content", ""),
                    completion_function_params,
                    _num_usages or 0,
                    _first_usage_date or "",
                )

                # Increment the number of usages for everything EXCEPT inline completions.
                if prompt_type != "inline_completion":
                    _num_usages = (_num_usages or 0) + 1
                    set_user_field(UJ_AI_MITO_API_NUM_USAGES, _num_usages)

                # Log the successful completion
                log_ai_completion_success(
                    key_type=MITO_SERVER_KEY,
                    prompt_type=prompt_type,
                    last_message_content=str(request.messages[-1].get('content', '')),
                    response={"completion": ai_response},
                    num_usages=_num_usages,
                )

                return CompletionReply(
                    parent_id=request.message_id,
                    items=[
                        CompletionItem(
                            content=ai_response,
                            isIncomplete=False,
                        )
                    ],
                )

        except BaseException as e:
            self.last_error = CompletionError.from_exception(e)
            key_type = MITO_SERVER_KEY if self.api_key is None else USER_KEY
            log(MITO_AI_COMPLETION_ERROR, params={KEY_TYPE_PARAM: key_type}, error=e)
            raise


    async def stream_completions(
        self, request: CompletionRequest, prompt_type: str, model: str
    ) -> AsyncGenerator[Union[CompletionReply, CompletionStreamChunk], None]:
        """Stream completions from the OpenAI API.

        Args:
            request: The completion request description.
            prompt_type: The type of prompt that was sent to the AI (e.g. "chat", "smart_debug", "explain")
        Returns:
            An async generator yielding first an acknowledge completion reply without
            completion and then completion chunks from the third-party provider.
        """
        # The streaming completion has two steps:
        # Step 1: Acknowledge the request
        # Step 2: Stream the completion chunks coming from the OpenAI API
        self.last_error = None

        # Acknowledge the request
        yield CompletionReply(
            items=[
                CompletionItem(content="", isIncomplete=True, token=request.message_id)
            ],
            parent_id=request.message_id,
        )
        
        # Validate that the model is supported. If not fall back to gpt-4o-mini
        if model not in self.models:
            model = "gpt-4o-mini"

        # Send the completion request to the OpenAI API and returns a stream of completion chunks
        try:
            completion_function_params = get_open_ai_completion_function_params(model, request.messages, stream=True)
            stream: AsyncStream[ChatCompletionChunk] = await self._openAI_async_client.chat.completions.create(**completion_function_params)
            
            # Log the successful completion
            log_ai_completion_success(
                key_type=USER_KEY,
                prompt_type=prompt_type,
                last_message_content=str(request.messages[-1].get('content', '')),
                response={"completion": "not available for streamed completions"},
            )
        except BaseException as e:
            self.last_error = CompletionError.from_exception(e)
            log(MITO_AI_COMPLETION_ERROR, params={KEY_TYPE_PARAM: USER_KEY}, error=e)
            raise

        async for chunk in stream:
            try:
                is_finished = chunk.choices[0].finish_reason is not None
                yield CompletionStreamChunk(
                    parent_id=request.message_id,
                    chunk=CompletionItem(
                        content=chunk.choices[0].delta.content or "",
                        isIncomplete=True,
                        token=request.message_id,
                    ),
                    done=is_finished,
                )
            except BaseException as e:
                self.last_error = CompletionError.from_exception(e)
                yield CompletionStreamChunk(
                    parent_id=request.message_id,
                    chunk=CompletionItem(
                        content="",
                        isIncomplete=True,
                        error=CompletionItemError(
                            message=f"Failed to parse chunk completion: {e!r}"
                        ),
                        token=request.message_id,
                    ),
                    done=True,
                    error=CompletionError.from_exception(e),
                )
                break



from __future__ import annotations

import traceback
from dataclasses import dataclass
from typing import List, Literal, Optional, Type

from pydantic import BaseModel
from openai.types.chat import ChatCompletionMessageParam

from .prompt_builders import (
    create_chat_prompt,
    create_inline_prompt,
    create_explain_code_prompt,
    create_error_prompt,
    create_agent_prompt,
)

CompletionIncomingMessageTypes = Literal['chat', 'inline_completion', 'codeExplain', 'smartDebug', 'agent:planning']
AllIncomingMessageTypes = Literal['clear_history', CompletionIncomingMessageTypes]

@dataclass(frozen=True)
class AICapabilities:
    """AI provider capabilities"""

    configuration: dict
    """Configuration schema."""
    provider: str
    """AI provider name."""
    type: str = "ai_capabilities"
    """Message type."""

@dataclass(frozen=True)
class ChatMessageBuilder:
    variables: Optional[List[str]] = None
    activeCellCode: Optional[str] = None
    input: Optional[str] = None
    index: Optional[int] = None

    @property
    def prompt(self) -> str:
        return create_chat_prompt(self.variables or [], self.activeCellCode or '', self.input or '')
    
    @property
    def pro_model(self) -> str:
        return "o3-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

@dataclass(frozen=True)
class SmartDebugMessageBuilder:
    variables: Optional[List[str]] = None
    activeCellCode: Optional[str] = None
    errorMessage: Optional[str] = None

    @property
    def prompt(self) -> str:
        return create_error_prompt(self.errorMessage or '', self.activeCellCode or '', self.variables or [])
    
    @property
    def pro_model(self) -> str:
        return "gpt-4o-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

@dataclass(frozen=True)
class CodeExplainMessageBuilder:
    variables: Optional[List[str]] = None
    activeCellCode: Optional[str] = None

    @property
    def prompt(self) -> str:
        return create_explain_code_prompt(self.activeCellCode or '')
    
    @property
    def pro_model(self) -> str:
        return "gpt-4o-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

@dataclass(frozen=True)
class InlineCompletionMessageBuilder:
    prefix: Optional[str] = None
    suffix: Optional[str] = None
    variables: Optional[List[str]] = None

    @property
    def prompt(self) -> str:
        return create_inline_prompt(self.prefix or '', self.suffix or '', self.variables or [])
    
    @property
    def pro_model(self) -> str:
        return "gpt-4o-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

@dataclass(frozen=True)
class AgentMessageBuilder:
    fileType: Optional[str] = None
    columnSamples: Optional[List[str]] = None
    input: Optional[str] = None

    @property
    def prompt(self) -> str:
        return create_agent_prompt(self.fileType or '', self.columnSamples or [], self.input or '')
    
    @property
    def pro_model(self) -> str:
        return "o3-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

    @property
    def response_format(self) -> Type[BaseModel]:
        class PlanOfAttack(BaseModel):
            actions: List[str]
            dependencies: List[str]
        return PlanOfAttack
    


@dataclass(frozen=True)
class CompletionRequest:
    """Message send by the client to request an AI chat response."""

    type: IncomingMessageTypes
    """Message type."""
    message_id: str
    """Message UID generated by the client."""
    messages: List[dict] = None
    """Chat messages."""
    stream: bool = False
    """Whether to stream the response (if supported by the model)."""


@dataclass(frozen=True)
class CompletionItemError:
    """Completion item error information."""

    message: Optional[str] = None
    """Error message."""


@dataclass(frozen=True)
class CompletionItem:
    """A completion suggestion."""

    content: str
    """The completion."""
    isIncomplete: Optional[bool] = None
    """Whether the completion is incomplete or not."""
    token: Optional[str] = None
    """Unique token identifying the completion request in the frontend."""
    error: Optional[CompletionItemError] = None
    """Error information for the completion item."""


@dataclass(frozen=True)
class CompletionError:
    """Completion error description"""

    error_type: str
    """Error type"""
    title: str
    """Error title"""
    traceback: str
    """Error traceback"""
    hint: str = ""
    """Hint to resolve the error"""

    @staticmethod
    def from_exception(exception: BaseException, hint: str = "") -> CompletionError:
        """Create a completion error from an exception."""
        error_type = type(exception)
        error_module = getattr(error_type, "__module__", "")
        return CompletionError(
            error_type=f"{error_module}.{error_type.__name__}"
            if error_module
            else error_type.__name__,
            title=exception.body.get("message")
            if hasattr(exception, "body")
            else (exception.args[0] if exception.args else "Exception"),
            traceback=traceback.format_exc(),
            hint=hint,
        )


@dataclass(frozen=True)
class ErrorMessage(CompletionError):
    """Error message."""

    type: Literal["error"] = "error"
    """Message type."""


@dataclass(frozen=True)
class CompletionReply:
    """Message sent from model to client with the completion suggestions."""

    items: List[CompletionItem]
    """List of completion items."""
    parent_id: str
    """Parent message UID."""
    type: Literal["reply"] = "reply"
    """Message type."""
    error: Optional[CompletionError] = None
    """Completion error."""


@dataclass(frozen=True)
class CompletionStreamChunk:
    """Message sent from model to client with the infill suggestions"""

    chunk: CompletionItem
    """Completion item."""
    parent_id: str
    """Parent message UID."""
    done: bool
    """Whether the completion is done or not."""
    type: Literal["chunk"] = "chunk"
    """Message type."""
    error: Optional[CompletionError] = None
    """Completion error."""



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
Utilities for creating and initializing the user.json
file with the current schema
"""

import json
import os

from .db import (MITO_FOLDER, USER_JSON_PATH, set_user_field)
from .schema import (GITHUB_ACTION_EMAIL, GITHUB_ACTION_ID,
                                    UJ_STATIC_USER_ID, UJ_USER_EMAIL,
                                    USER_JSON_DEFAULT)
from .telemetry_utils import identify
from .utils import is_running_test


def is_user_json_exists_and_valid_json() -> bool:
    """
    Helper function that determines if the current user.json both
    exists and is valid json
    """
    if not os.path.exists(USER_JSON_PATH):
        return False

    try:
        with open(USER_JSON_PATH, 'r') as f:
            json.loads(f.read())
        return True
    except:
        return False


def try_create_user_json_file() -> None:
    
    # Create the mito folder if it does not exist
    if not os.path.exists(MITO_FOLDER):
        os.mkdir(MITO_FOLDER)

    # We create a user.json file if it does not exist, or if it
    # is invalid (e.g. it is not parseable JSON).
    if not is_user_json_exists_and_valid_json():
        # First, we write an empty default object
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(USER_JSON_DEFAULT))

        # Then, we take special care to put all the testing/CI environments 
        # (e.g. Github actions) under one ID and email
        if is_running_test():
            set_user_field(UJ_STATIC_USER_ID, GITHUB_ACTION_ID)
            set_user_field(UJ_USER_EMAIL, GITHUB_ACTION_EMAIL)


def initialize_user() -> None:
    """
    Internal helper function that gets called whenever a ai completion is requested.

    It:
    1. Creates the user.json if it does not exist (though it usually does, from the installer)
    2. Identifies the user
    """
    # Try to create the user.json file, if it does not already exist
    try_create_user_json_file()

    # Identify the user 
    identify()

    # Note, its possible that a user has a previous version of the user.json file if they 
    # downloaded Mito >1 year ago and have not yet upgraded. In this case, we would like to 
    # upgrade the user.json to the newest versio. But for simplicity, let's not 
    # try to upgrade it here. We will either move away from user.json all together
    # or unify the utilities for mitosheet and mitoai.


from glob import glob
from typing import List, Tuple
from setuptools import setup, find_packages
import json
from pathlib import Path
import os

HERE = Path(__file__).parent.resolve()
package_json = json.loads(open('package.json').read())
lab_path = Path(HERE, 'mito_ai', 'labextension')


data_files_spec = [
    # Lab extension data files
    ("share/jupyter/labextensions/mito-ai", str(lab_path), "**"),
    ("share/jupyter/labextensions/mito-ai", str(HERE), "install.json")
]

def get_data_files_from_data_files_spec(
    data_specs: List[Tuple[str, str, str]],
):
    """
    Given tuples of (data_file_path, directory_to_search, pattern_to_find),
    this function will return a list of tuples of (data_file_path, [files])
    in the format that setuptools expects.
    """
    file_data = {}
    for (data_file_path, directory_to_search, pattern_to_find) in data_specs or []:
        # Get the directory to search ready
        if os.path.isabs(directory_to_search):
            directory_to_search = os.path.relpath(directory_to_search)
        directory_to_search = directory_to_search.rstrip("/")
        # Get all non-directory files that match the pattern, searching recursively
        files = [
            f for f in glob(
                Path().joinpath(directory_to_search, pattern_to_find).as_posix(), 
                recursive=True
            ) if not os.path.isdir(f)
        ]
        offset = len(directory_to_search) + 1
        for file_name in files:
            relative_path = str(Path(file_name).parent)[offset:]
            full_data_file_path = Path().joinpath(data_file_path, relative_path).as_posix()
            if full_data_file_path not in file_data:
                file_data[full_data_file_path] = []
            file_data[full_data_file_path].append(file_name)
    # Turn to list and sort by length, to be consistent (and maybe cuz we need to for folder creation?)
    data_files = sorted(file_data.items(), key=lambda x: len(x[0]))
    
    return data_files

data_files = get_data_files_from_data_files_spec(data_files_spec) 

# Add the jupyter server config file so that the extension is automatically loaded
data_files.append(("etc/jupyter/jupyter_server_config.d", ["jupyter-config/jupyter_server_config.d/mito-ai.json"]))

# Read the content of README.md for the long description
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="mito-ai",
    version=package_json['version'],
    author="Aaron Diamond-Reivich",
    author_email="aaron@sagacollab.com",
    description="AI chat for JupyterLab",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GNU Affero General Public License v3",
    python_requires=">=3.9",
    classifiers=[
        "Framework :: Jupyter",
        "Framework :: Jupyter :: JupyterLab",
        "Framework :: Jupyter :: JupyterLab :: 4",
        "Framework :: Jupyter :: JupyterLab :: Extensions",
        "Framework :: Jupyter :: JupyterLab :: Extensions :: Prebuilt",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    packages=find_packages(),
    install_requires=[
        "jinja2>=3.0.3",
        "jupyterlab>=4.0.0,<5",
        "openai>=1.0.0",
        'analytics-python',
        "tornado>=6.2.0",
        "traitlets",
    ],
    extras_require = {
        'deploy': [
            'wheel==0.42.0', 
            'twine==5.1.1',
            'setuptools==68.0.0'
            
        ],
        'test': [
            'pytest==8.3.4',
        ],
    },
    keywords=["AI", "Jupyter", "Mito"],
    entry_points={
        "jupyter_serverproxy_servers": [
            "mito-ai = mito_ai:_load_jupyter_server_extension",
        ],
    },
    jupyter_server_extension="mito_ai", # Automatically enable the server extension 
    include_package_data=True,  # Ensures labextension files are included
    package_data={
        "": ["labextension/**/*"],
    },
    data_files=data_files,
    zip_safe=False,
)